/*
 * unix-ex.c: example of server using Unix domain sockets.
 *
 * This is an excerpt of fieldd.c (a magnetic field server) with all the
 * code relative to field control removed. It is therefore competely
 * useless exept as an example or skeleton for writing another server:
 * look for the string "many other commands" and put the commands
 * specific to your server there. Below are the original comments at the
 * beginning of fieldd.c. Some of the them are not relevant to this
 * excerpt: in this example
 *      - the math library is not used
 *      - the macros ECHO_COMMANDS and HISTOGRAM have no effect
 *      - it doesn't need root privileges.
 *
 * ---------------------------------------------------------------------
 *
 * fieldd: field daemon.
 *
 * This program controls the field applied by the magnets. At
 * initialization it opens a Unix socket in /tmp/fieldd-socket (but see
 * below for overriding) and listens for ASCII-formatted commands sent
 * over it. Look at the `help' variable for available commands. When
 * sweeping the field, it updates the current in the magnet every
 * millisecond. Field data is sent serialized through the fisrst
 * parallel port (parport0).
 *
 * This program has to be linked to the math library (compile with -lm).
 * The following preprocessor macros can be used to alter it's
 * behaviour:
 *      SOCKET_NAME
 *          If this macro is set, it's value will override the default
 *          socket name ("/tmp/fieldd-socket").
 *      ECHO_COMMANDS:
 *          If this is set, fieldd will echo all received commands on
 *          sterr. It wil also display a rotating bar (alternating -, \,
 *          | and /) on stderr when sweeping the field.
 *      HISTOGRAM:
 *          If this is set, fieldd will compute an histogram of delays
 *          between calls to update_field() and write it to
 *          "histogram.tsv" upon program termination. The bins are
 *          evenly spaced on a logarithmic scale.
 *
 * This program has to be installed suid root in order to gain access to
 * the I/O space of the parallel port and to get real-time scheduling
 * priority. It will drop root privileges during initialization.
 *
 * TODO:
 *    - Zero the field before quitting.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#ifndef SOCKET_NAME
# define SOCKET_NAME "/tmp/fieldd-socket"
#endif

#define MAX_CLIENTS 5           /* max number of network clients */


/***********************************************************************
 * I/O processing.
 *
 * To avoid blocking on input, we do a single read() when select() tells
 * us it will not block.
 *
 * To avoid blocking on output, we write everything on an internal
 * buffer. When there is data in the buffer, we do a single write() when
 * selects() tells us it will not block.
 */

typedef struct {
    int active;                 /* 1 if this slot is used */
    int s;                      /* connected socket */
    size_t output_pending;      /* number of bytes pending */
    char output_buffer[4096];
} client_t;

client_t client[MAX_CLIENTS];

client_t *get_client_slot(void)
{
    int i;

    for (i=0; i<MAX_CLIENTS; i++)
        if (!client[i].active) return &client[i];
    return NULL;
}

/* Cleanup, this has to be registered with atexit(). */
void unlink_socket(void) { unlink(SOCKET_NAME); }

/*
 * Get a listen()ing socket ready to accept() a connection.
 * Returns -1 on error.
 */
int get_socket(void)
{
    int s;
    struct sockaddr_un my_addr;
    int err;

    /* Get a socket. */
    s = socket(PF_UNIX, SOCK_STREAM, 0);
    if (s == -1) { perror("socket"); return -1; }

    /* bind() it. */
    my_addr.sun_family = AF_UNIX;
    strcpy(my_addr.sun_path, SOCKET_NAME);
    err = bind(s, (struct sockaddr *) &my_addr, sizeof my_addr);
    if (err == -1) { perror("bind"); return -1; }

    /* We want to unlink the named socket at program termination. */
    err = atexit(unlink_socket);
    if (err) {
        fprintf(stderr, "atexit failed\n");
        unlink_socket();
        return -1;
    }

    /* listen() to one client at a time. */
    err = listen(s, 1);
    if (err) { perror("listen"); return -1; }

    return s;
}

void process_output(client_t *cl)
{
    int ret;

    if (!cl->output_pending) return;    /* be defensive */
    ret = write(cl->s, cl->output_buffer, cl->output_pending);
    if (ret < 0) { perror("fieldd: write"); return; }
    if (ret > 0 && (unsigned) ret < cl->output_pending) {
        memmove(cl->output_buffer, cl->output_buffer + ret,
                cl->output_pending - ret);
    }
    cl->output_pending -= ret;
}

void queue_output(client_t *cl, char *fmt, ...)
{
    va_list ap;
    int available;
    int n;

    available = sizeof cl->output_buffer - cl->output_pending;
    va_start(ap, fmt);
    n = vsnprintf(cl->output_buffer + cl->output_pending, available,
            fmt, ap);
    va_end(ap);

    if (n >= available) {
        fprintf(stderr, "Output buffer overflow\n");
        exit(EXIT_FAILURE);
    }
    cl->output_pending += n;
}


/***********************************************************************
 * Command interpreter.
 */

int should_quit = 0;

/* Available commands. */
char *help =
"Available commands:\n"
"help?           return command summary\n"
"...             many other commands...\n"
"quit            close connection\n"
"exit            stop the server\n"
;

/*
 * Returns the return value of read().
 */
int process_input(client_t *cl)
{
    char buffer[1024];
    int ret;
    char *delimiters = " \t,\n";
    char *token;

    /* Read the command line. */
    ret = read(cl->s, buffer, sizeof buffer - 1);
    if (ret == -1) {
        perror("fieldd: read");
        exit(EXIT_FAILURE);
    }
    buffer[ret] = '\0';

    /* End of file? */
    if (ret == 0) return ret;

    /* Get the first token (the command). */
    token = strtok(buffer, delimiters);
    if (!token) return ret;

    /* Respond accordingly... */

    if (strcmp(token, "help") == 0 || strcmp(token, "help?") == 0) {
        queue_output(cl, help);
        return ret;
    }

    /* ... many other commands... some of them requiring extra tokens. */

    if (strcmp(token, "quit") == 0) {
        close(cl->s);
        return 0;       /* equivalent to an end of file */
    }

    if (strcmp(token, "exit") == 0) {
        should_quit = 1;
        return ret;
    }

    fprintf(stderr, "Command unrecognized\n");
    return ret;
}


/***********************************************************************
 * Main: initialization and select() loop.
 */

#define FD_SET_M(fd, set) do { FD_SET(fd, set); \
        max_fd = fd>max_fd ? fd : max_fd; } while (0)

int main(int argc, char *argv[])
{
    int i;
    client_t *cl;
    int ls;                         /* listening socket */
    struct sockaddr_un peer_addr;
    int peer_lg;
    fd_set rfds, wfds;
    int max_fd;
    int ret;

    (void) argc; (void) argv;

    /* Get a listening socket. */
    ls = get_socket();
    if (ls == -1) return EXIT_FAILURE;

    do {

        /* select() loop. */
        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        max_fd = 0;
        for (i=0; i<MAX_CLIENTS; i++) {
            cl = &client[i];
            if (cl->active) {
                FD_SET_M(cl->s, &rfds);
                if (cl->output_pending)
                    FD_SET_M(cl->s, &wfds);
            }
        }
        cl = get_client_slot();
        if (cl) FD_SET_M(ls, &rfds);
        ret = select(max_fd + 1, &rfds, &wfds, NULL, NULL);
        if (ret == -1) { perror("select"); return EXIT_FAILURE; }

        /* accept() connection. */
        if (FD_ISSET(ls, &rfds)) {
            peer_lg = sizeof peer_addr;
            cl->s = accept(ls, (struct sockaddr *) &peer_addr, &peer_lg);
            if (cl->s == -1) {
                perror("accept");
                return EXIT_FAILURE;
            }
            cl->active++;
            cl->output_pending = 0;
        }

        /* Do I/O. */
        for (i=0; i<MAX_CLIENTS; i++) {
            cl = &client[i];
            if (cl->s > max_fd) continue;   /* this is the new client */
            if (FD_ISSET(cl->s, &wfds)) process_output(cl);
            if (FD_ISSET(cl->s, &rfds))
                if(process_input(cl) == 0) {
                    cl->active = 0;         /* client disconnected */
                }
        }

    } while (!should_quit);

    return EXIT_SUCCESS;
}
